#include<bonnet.h>

/*************************************
*
*        Free list Function
*
**************************************/

static struct buffer_head *__get_freelist()
{
	return free_list->next;
}

/* Node Allocation */
static struct buffer_head *__nalloc()
{
	struct buffer_head *ret;
	ret = (struct buffer_head *)__get_freelist();

	if(ret == NULL)
		return NULL;

	if(ret->prev) ret->prev->next = ret->next;
	if(ret->next) ret->next->prev = ret->prev;

	return ret;
}

/* Node Free */
static void __nfree(struct buffer_head *tmp)
{
	if(tmp->prev) tmp->prev->next = tmp->next;
	if(tmp->next) tmp->next->prev = tmp->prev;
}

/*************************************
 *
 * 		Hash list Function
 *
 *************************************/

static int _nalloc(struct buffer_head *new)
{
	struct buffer_head *start;
	struct buffer_head *tmp;

	start = hash_list[new->blkno % HASHKEY];
	if(start->next == NULL)
	{
		start->next = new;
		new->prev = start;
		new->next = NULL;
	}
	else
	{
		tmp = start->next;
		start->next = new;

		new->prev = start;
		new->next = tmp;

		tmp->prev = new;
	}
	return 1;
}

static int _nfree(int blkno)
{
	struct buffer_head *del;

	del = (struct buffer_head *)_nfind(blkno);

	if(del)
	{
		if(del->prev) del->prev->next = del->next;
		if(del->next) del->next->prev = del->prev;

		del->prev = del->next = NULL;
	}
	return 1;
}

struct buffer_head *_nfind(int blkno)
{
	struct buffer_head *ret;
	int hashkey;

	hashkey = blkno % HASHKEY;
	ret = hash_list[hashkey];
	
	while(ret && ret->next && (ret = ret->next) != NULL)
	{
		if(ret->blkno == blkno)
			return ret;	
	}
	return NULL;
}

/*************************************
 *
 *		  Priority Fuction
 *
 **************************************/
struct buffer_head *select_low_priority()
{
	return hash_list[last_alloc_hash]->next;
}

void plus_priority(struct buffer_head *tmp)
{
	struct buffer_head *buf;

	if(tmp->next)
	{
		buf = tmp->next;

		buf->prev = tmp->prev;
		tmp->next = buf->next;

		buf->next = tmp;
		tmp->prev = buf;
	}
}

/*************************************
 *
 * 		     Main Function
 *
 *************************************/

// insert to buffer
// 1. delete from free_list
// 2. alloc hash_list
int insert_buffer(int blkno,char *buf)
{
	struct buffer_head *new;

	new = __nalloc();
	if(new == NULL)
	{
		// when it has no free buffer, delete other
		new = select_low_priority();
		if(new == NULL) return 0;
		delete_buffer(new); // delete it!
	}

	new->status = STATUS_ALLOC;
	new->count  = 1;
	new->blkno = blkno;
	if(new->buffer == NULL) new->buffer = (char *)malloc(BLKSIZE);
	memcpy(new->buffer,buf,BLKSIZE);

	// insert to hash_list
	_nalloc(new);
}

int delete_buffer(struct buffer_head *tmp)
{
	_nfree(tmp->blkno);
	__nfree(tmp);
}

// find the block
struct buffer_head *find_buffer(int blkno)
{
	return _nfind(blkno);
}

// read from buffer
int getblk(int blkno,char *buf,int count)
{
	struct buffer_head *tmp;
	char *ptr = buf;
	int i;

	if(blkno < 0)
		return 0;

	if(count <= 0)
		return 0;

	for(i=0;i<count;i++)
	{
		tmp = find_buffer(blkno+i);

		// if buffer has available,..
		if(tmp != NULL)
		{
			plus_priority(tmp);
			memcpy(ptr,tmp->buffer,BLKSIZE);	
		}
		// if not, access to native
		else
		{
			bread(blkno+i,ptr,1);
			insert_buffer(blkno+i,ptr);
		}
		ptr+=BLKSIZE;
	}
	return 0;
}

static void init_free_list()
{
	struct buffer_head *node = NULL;
	struct buffer_head *prev = NULL;
	int i;

	LIST_INIT(free_list);
	prev = free_list;

	for(i=0;i<MAX_BUFFER;i++)
	{
		LIST_INIT(node);
		
		// setting PREV,NEXT
		if(node != NULL) 
		{
			prev->next = node;
			node->prev = prev;
		}

		prev = node;
	}
}

static void init_hash_list()
{
	int i;

	for(i=0;i<MAX_HASH;i++)
		LIST_INIT(hash_list[i]);
}

void init_buffer_cache()
{
	init_free_list();
	init_hash_list();
}
